//@line 4 "/builds/slave/m-b30_14-l64_g-ntly-0000000000/build/b2g/locales/en-US/b2g-l10n.js"

//@line 6 "/builds/slave/m-b30_14-l64_g-ntly-0000000000/build/b2g/locales/en-US/b2g-l10n.js"

pref("general.useragent.locale", "en-US");
